/****************************************************** 
 * File  : action.cc   
 * Desc. : functions for extracting actions from 
 *         universal plan
 * Author: Rune M. Jensen
 * Date  : 5/7/01
 ******************************************************/
#include <bdd.h>
#include <stream.h>
#include <math.h>
#include "execute.h"
#include "action.h"


bdd plan;

int en[6][3] =  // BDD var encoding of current vars; 
{ {8, 10,12},  // o1x
  {14,16,18},  // o2x
  {20,22,24},  // Vx
  {28,30,32},  // o1y
  {34,36,38},  // o2y
  {40,42,44}   // Hy
};

int Vact[4] = {0,1,2,3};
int Hact[4] = {4,5,6,7};
 

 


void InitPlan(char *planname) {

  // initialize bdd package 
  bdd_init(100000,40000);
  bdd_setvarnum(100);
  
  
  // read bdd infile
  if (bdd_fnload(planname,plan))  {
    cout << "action.cc : InitPlan : Cannot load bdd from file \"" << planname << "\"\nexiting\n";
    exit(1);
  }
}



action ChooseAct(gridstate st) {

  bdd actbdd;
  action res;

  // constrain universal plan by current state
  actbdd = plan;
  actbdd &= gridstate2bdd(st);


  bdd nopnop = int2bdd(Vact,4,VNop) & int2bdd(Hact,4,HNop);
  bdd Vnop   = int2bdd(Vact,4,VNop);
  bdd Hnop   = int2bdd(Hact,4,HNop);

  if ( (actbdd & nopnop ) != bddfalse ) {
    // both nop actions
    res.Ver = VNop;
    res.Hor = HNop;
  }
  else if ( (actbdd & Vnop ) != bddfalse ) {
    // vertical is nop
    res.Ver = VNop;
    res.Hor = (enum HorAct) NoOfAct(Hact,4,bdd_satone(actbdd & Vnop));
  }
  else if ( (actbdd & Hnop ) != bddfalse ) {
    // horizontal is nop
    res.Ver = (enum VerAct) NoOfAct(Vact,4,bdd_satone(actbdd & Hnop));
    res.Hor = HNop;
  }
  else { // none nop
    actbdd = bdd_satone(actbdd);
    res.Ver = (enum VerAct) NoOfAct(Vact,4,actbdd);
    res.Hor = (enum HorAct)NoOfAct(Hact,4,actbdd);
  }

  return res;
}




//IN
// *var   : bdd variable mapping
// length : length of mapping
//OUT
// natural number represented by BDD variables
int NoOfAct(int *var,int length, bdd a) {

  int i;
  int b[16];
  bdd d;

  // extract action number in binary
  // obs: bit_0 most significant
  for (i=0; i < length; i++)
    {
      d = bdd_ithvar(var[i]);
      d &= a;
      if (d == bddfalse)
	b[i] = 0;
      else
	b[i] = 1;	
    }

  // convert to int
  int exp,res;
  
  res = 0;
  exp =  length - 1;
  for (i= 0; i < length; i++)
    res += b[i]*int(pow(2,exp--));

  //return action value
  return res;
}






bdd gridstate2bdd(gridstate st) {
  
  bdd res = bddtrue;

  res &= int2bdd(en[0],3,st.o1.x);
  res &= int2bdd(en[1],3,st.o2.x);
  res &= int2bdd(en[2],3,st.v.x);
  res &= int2bdd(en[3],3,st.o1.y);
  res &= int2bdd(en[4],3,st.o2.y);
  res &= int2bdd(en[5],3,st.h.y);
  
  if (st.von)
    res &= bdd_ithvar(26); 
  else
    res &= bdd_nithvar(26);
    
  if (st.hon)
    res &= bdd_ithvar(46); 
  else
    res &= bdd_nithvar(46);

  return res;
}
  

// IN
//   var    : bdd var mapping
//   length : number of bdd vars
//   n      : integer to encode.
//   OUT
//   bdd : BDD with the binary encoding of n
//         starting with the most significant bit. 
//         so if you write out the bdd vars 012345 the binary will
//         have its least sig. bit leftmost !!
//         All numbers are encoded by UMOP like this (also act. vars)
bdd int2bdd(int *var,int length,int n)  {
  bdd res;
  int j;  

  res = bddtrue;
  for (j=length-1; j >= 0; j--)
    {
      if (n % 2) 
	res &= bdd_ithvar(var[j]); 
      else
	res &= bdd_nithvar(var[j]);
      n /= 2;
    }
return res;
}



