/****************************************************** 
 * File  : execute.cc   
 * Desc. : execution module
 * Author: Rune M. Jensen
 * Date  : 5/7/01
 ******************************************************/
#include <X11/Xlib.h>
#include <stdlib.h>
#include "main.h"
#include "draw.h"
#include "action.h"
#include "execute.h"



// In 
// st          : state to execute action in
// act         : action to execute 
// speed       : simulation speed (1,2,...)
// screen_size : screen_size (1,2,...)
// Out
//  resulting state (side effect visual execution of joint action)
gridstate execute(DisplayInfo di, gridstate old_st, action act, int speed) {
  
  gridstate new_st;

  // compute new state
  new_st = old_st;
  switch (act.Ver) {  
  case VRightOn12 :
    new_st.o1.x++;
    new_st.o2.x++;
    new_st.v.x++;
    break;
    
  case VRightOn1 :
    new_st.o1.x++;
    new_st.v.x++;
    break;
    
  case VRightOn2 :
    new_st.o2.x++;
    new_st.v.x++;
    break;
    
  case VRight :
    new_st.v.x++;
    break;
    
  case VLeftOn12 :
    new_st.o1.x--;
    new_st.o2.x--;
    new_st.v.x--;
    break;
    
  case VLeftOn1 :
    new_st.o1.x--;
    new_st.v.x--;
    break;
    
  case VLeftOn2 :
    new_st.o2.x--;
    new_st.v.x--;
    break;
    
  case VLeft :
    new_st.v.x--;
    break;
    
  case VFlip :
    new_st.von = ! new_st.von; 
    break;
    
  case VNop :
    break;
  }
  
  
  switch (act.Hor) {  
  case HUpOn12 :
    new_st.o1.y++;
    new_st.o2.y++;
    new_st.h.y++;    
    break;
    
  case HUpOn1 :
    new_st.o1.y++;
    new_st.h.y++;    
    break;
    
  case HUpOn2 :
    new_st.o2.y++;
    new_st.h.y++;    
    break;
    
  case HUp :
    new_st.h.y++;    
    break;
    
  case HDownOn12 :
    new_st.o1.y--;
    new_st.o2.y--;
    new_st.h.y--;    
    break;
    
  case HDownOn1 :
    new_st.o1.y--;
    new_st.h.y--;    
    break;
    
  case HDownOn2 :
    new_st.o2.y--;
    new_st.h.y--;    
    break;
    
  case HDown :
    new_st.h.y--;    
    break;
    
  case HFlip :
    new_st.hon = ! new_st.hon;    
    break;
    
  case HNop :
	break;
  }

  // simulate change
  Simulate(di,old_st,new_st,speed);
  

  return new_st;
}



gridstate initstate(void) {
  
  gridstate res;
  
  res.o1.y = rand() % 8;
  res.o2.y = rand() % 8;

  int x1 = rand() % 8;
  int x2 = rand() % 8;
  while ( (x2 = rand() % 8) == x1 );
  if (x1 > x2) {
    res.o1.x = x1;
    res.o2.x = x2;
  }
  else {
    res.o1.x = x2;
    res.o2.x = x1;
  }
    
  res.v.x = rand() % 8;
  res.v.y = 7;
  res.h.x = 0;
  res.h.y = rand() % 8;
  
  res.von = false;
  res.hon = false;

  return res;
}
    

gridstate newinitstate(gridstate res) {
    
  res.o1.y = rand() % 8;
  res.o2.y = rand() % 8;

  int x1 = rand() % 8;
  int x2 = rand() % 8;
  while ( (x2 = rand() % 8) == x1 );
  if (x1 > x2) {
    res.o1.x = x1;
    res.o2.x = x2;
  }
  else {
    res.o1.x = x2;
    res.o2.x = x1;
  }
    
  return res;
}
    

bool goal_reached(gridstate st) {
  return (st.o1.x == 7 && st.o1.y == 7 && st.o2.x == 6 && st.o2.y == 7);
}


void Move2GridState(DisplayInfo di,gridstate st) {

  screenstate sst;

  sst.o1 = Grid2ScreenPos(st.o1,di.screen_size);
  sst.o2 = Grid2ScreenPos(st.o2,di.screen_size);
  sst.v = Grid2ScreenPos(st.v,di.screen_size);
  sst.h = Grid2ScreenPos(st.h,di.screen_size);

  // move things around
  
  XMoveWindow(di.display,di.Obj1.win,int(sst.o1.x + di.screen_size/120),int(sst.o1.y + di.screen_size/120));
  XMoveWindow(di.display,di.Obj2.win,int(sst.o2.x + di.screen_size/120),int(sst.o2.y + di.screen_size/120));
  XMoveWindow(di.display,di.BarH.win,int(sst.h.x),int(sst.h.y + 3*di.screen_size/80));
  XMoveWindow(di.display,di.BarV.win,int(sst.v.x + 3*di.screen_size/80),int(sst.v.y));


  DrawRec(di.display,di.Obj1,"gold");
  DrawRec(di.display,di.Obj2,"RoyalBlue");


  // change bar color 
  if (st.hon)
    DrawRec(di.display,di.BarH,"red3");
  else
    DrawRec(di.display,di.BarH,"black");

  if (st.von)
    DrawRec(di.display,di.BarV,"red3");
  else
    DrawRec(di.display,di.BarV,"black");

  /* redraw grid */
  DrawGrid(di.display,di.grid,di.screen_size);

  XFlush(di.display);

}
  


void Simulate(DisplayInfo di,gridstate old_st,gridstate new_st,int speed) {

  screenstate old_sst;
  screenstate d_sst;

  old_sst.o1 = Grid2ScreenPos(old_st.o1,di.screen_size);
  old_sst.o2 = Grid2ScreenPos(old_st.o2,di.screen_size);
  old_sst.v = Grid2ScreenPos(old_st.v,di.screen_size);
  old_sst.h = Grid2ScreenPos(old_st.h,di.screen_size);

  d_sst.o1 = DeltaScreenPos(old_st.o1,new_st.o1,di.screen_size);
  d_sst.o2 = DeltaScreenPos(old_st.o2,new_st.o2,di.screen_size);
  d_sst.v = DeltaScreenPos(old_st.v,new_st.v,di.screen_size);
  d_sst.h = DeltaScreenPos(old_st.h,new_st.h,di.screen_size);

  
  // move things around
  for (int i = 1; i < di.screen_size/(10*speed); i = i + speed) {

    XMoveWindow(di.display,di.Obj1.win,int(old_sst.o1.x + i*d_sst.o1.x + di.screen_size/120),
		int(old_sst.o1.y + i*d_sst.o1.y + di.screen_size/120));
    XMoveWindow(di.display,di.Obj2.win,int(old_sst.o2.x + i*d_sst.o2.x + di.screen_size/120),
		int(old_sst.o2.y + i*d_sst.o2.y + di.screen_size/120));
    XMoveWindow(di.display,di.BarH.win,int(old_sst.h.x + i*d_sst.h.x),
		int(old_sst.h.y + i*d_sst.h.y + 3*di.screen_size/80));
    XMoveWindow(di.display,di.BarV.win,int(old_sst.v.x + i*d_sst.v.x + 3*di.screen_size/80),
		int(old_sst.v.y + i*d_sst.v.y));

    DrawRec(di.display,di.Obj1,"gold");
    DrawRec(di.display,di.Obj2,"RoyalBlue");
    
    // change bar color 
    if (new_st.hon)
      DrawRec(di.display,di.BarH,"red3");
    else
      DrawRec(di.display,di.BarH,"black");
    
    if (new_st.von)
      DrawRec(di.display,di.BarV,"red3");
    else
      DrawRec(di.display,di.BarV,"black");


    

    /* draw grid */
    DrawGrid(di.display,di.grid,di.screen_size);
    
    XFlush(di.display);
    
  }
}

 


screenpos Grid2ScreenPos(gridpos gp,int screen_size) {

  screenpos res;

  res.x = (1 + gp.x) * screen_size/10;
  res.y = screen_size - (2 + gp.y)*screen_size/10 ;

  return res;
}


screenpos DeltaScreenPos(gridpos old_p, gridpos new_p, int screen_size) {

  screenpos n_p;
  screenpos o_p;
  screenpos res;
  n_p = Grid2ScreenPos(new_p,screen_size);
  o_p = Grid2ScreenPos(old_p,screen_size);

  res.x = double(n_p.x - o_p.x) / double(screen_size/10);
  res.y = double(n_p.y - o_p.y) / double(screen_size/10);
  
  return res;
}



