/*========================================================================
                              
                                ** UMOP Executor **

            Copyright (C) 1999 by Rune M. Jensen and Manuela M. Veloso
            Computer Science Department, Carnegie Mellon University.
            
    Permission is hereby granted to distribute the UMOP software for 
    research purposes, provided this notice is included with any such 
    distribution. 

      *** WE ARE NOT RESPONSIBLE FOR ANY KIND OF DAMAGE TO YOUR  ***
      *** FILES, DATA, HARDWARE, LOSS OF MONEY, SYSTEM CRASHES   *** 
      *** OR ANY OTHER THING YOU MIGHT COME UP WITH.             ***
      *** - USE THIS PROGRAM OF YOUR OWN FREE WILL !!!           ***

========================================================================*/
/****************************************************** 
 * File  : main.cc   
 * Desc. : main file for the UMOP executor
 * Author: Rune M. Jensen
 * Date  : 10/16/01
 ******************************************************/
#include <stream.h>
#include <string.h>
#include <stdio.h> 
#include <stdlib.h>
#include <bdd.h>
#include "common.hpp"
#include "domain.hpp"
#include "fsm.hpp"
#include "exefunc.hpp" // implementation of specific executor functions
#include "bddprint.hpp"
#include "main.hpp"


void  printusage();

int verbosity;

int main(int argc,char **argv)  {
  extern FILE *yyin;
  FILE *trace;
  domdef *def;
  
  int i,ind;
  char buf[16];
  int res;

  // default values of input parameters
  char domainname[MAXNAMELENGTH] = "domain.nadl";
  char planname[MAXNAMELENGTH] =   "universal.plan" ;
  char tracename[MAXNAMELENGTH] =   "trace.txt" ;

  int mktrace    = 0;
  int dynreorder = 0;
  int bddnodenum = 100000;
  int cachesize  = 10000;
  int partlimit  = 5000;
  int sysmanual  = 0;
  int envmanual  = 0;

 
  verbosity      = 0;
		    
  if (argc < 2) {
    printusage();
    exit (1);
  }

  cout << "Parsing NADL domain...\n";
  cout.flush();

  /*** Scan command line ***/
  for (ind = 1; ind < argc; ind++) {
    
    /*** Parse switches ***/
    if (argv[ind][0] == '-') {
      switch (argv[ind][1]) {  
      case 'd': strcpy(domainname,argv[++ind]);
	break;
      case 'i': strcpy(planname,argv[++ind]);
	break;
      case 'o': strcpy(tracename,argv[++ind]);
	mktrace = 1;
	break;
      case 'n': bddnodenum = atoi(argv[++ind]);
	break;
      case 'c': cachesize  = atoi(argv[++ind]);
	break;
      case 'p': partlimit  = atoi(argv[++ind]);
	break;
      case 'v': verbosity  = atoi(argv[++ind]);
	break;
      case 's': sysmanual  = 1;
	break;
      case 'e': envmanual  = 1;
	break;       
      case 'h': printusage();
	exit(0);
	break;
      default : printf("main.cc main : Unknown switch '%c'\nexiting\n", argv[ind][1]);
	exit(1);
	break;
      }
    }
  }


  // read output file
  if (mktrace)
    {
      trace = fopen(tracename,"w");
      if (trace == NULL) 
	{
	  cout << "main.cc main : Cannot open \"" << tracename << "\"\n";
	  exit(1);
	}
    }


  // parse umop domain
  yyin = fopen(domainname,"r");
  if (yyin == NULL) 
    {
      cout << "main.cc main : Cannot open \"" << domainname << "\"\n";
      exit(1);
    }

  // exit on syntax error
  if (yyparse()) exit(1); 

  if (verbosity > 1)
    prob->print();
  
  // make domain info. struct (1st add of info to def)
  def = mkdomdef(prob);

  if (verbosity > 1)
    def->print();
    
  // initialize bdd package
  bdd_init(bddnodenum,cachesize);
  bdd_setvarnum(def->bddvarnum);
        
  // add bdd structures defining variable locations
  def->rels = mkvarrels(def); //(2nd add of info to def)

  cout << "Reading plan...\n";  

  bdd plan,init,goal;
  
  // read plan
  if (bdd_fnload(planname,plan))  {
    cout << "action.cc : InitPlan : Cannot load bdd from file \"" << planname << "\"\nexiting\n";
    exit(1);
  }

  init = formula2bdd(prob->init,def);
  goal = formula2bdd(prob->goal,def);

  cout.flush();


  cout << "Building T...\n";  
  trel *T; 
 
  T = mkpartitions(def,false);
  ranjanoptimize(T,partlimit); 
  

  // debug
  //bdd t;
  //  t = bddtrue;
  //for (int p = 0; p < T->partnum; p++)
  //  t &= T->p[p].exp;
  //
  //bddprintabs pr("t.bdd",def);
  //pr.bddinfo(T->p[0].exp);
  //exit(1);


  // main loop

  bdd currentstate;
  int goprg = 1;
  int traceno = 1;
  
  
  while (goprg) 
    {
      printf("\nTrace %d\n",traceno);
      if (mktrace) fprintf(trace,"\nTrace %d\n",traceno);
      traceno++;

      currentstate = chooseInit(def,init,mktrace,trace);
      
      int gotrace = 1;
      int stepno = 1;

      while (gotrace)
	{
	  printf("\n  Step %d\n",stepno);
	  if (mktrace) fprintf(trace,"\n  Step %d\n",stepno);
	  stepno++;
	  
	  bdd sysact;
	  bdd envact;
	  
	  sysact = chooseSysAct(def,T,plan,currentstate,sysmanual,mktrace,trace);
	  if (sysact == bddfalse) gotrace = 0; 
	  else
	    {
	      envact = chooseEnvAct(def,T,plan,currentstate,envmanual,mktrace,trace);
	      if (envact == bddfalse) gotrace = 0;
	      else
		{
		  
		  // choose and print new state
		  currentstate = chooseNewState(def,envact,sysact,T,currentstate,mktrace,trace);

		  if ( ( currentstate & goal ) == currentstate )
		    { 
		      printf("\n  Trace reached a goal state !\n");
		      if (mktrace) fprintf(trace,"\n  Trace reached a goal state !\n");

		      gotrace = 0;
		    }
		  else
		    {
		      printf("\n  Continue current trace ? [y/n] : ");
		      scanf("%s",buf);
		      if (buf[0] == 'n') gotrace = 0;
		    }
		}
	    }
	}
      printf("\n  make a new trace ? [y/n] : ");
      scanf("%s",buf);
      if (buf[0] == 'n') goprg = 0;
    }
  
  // close bdd package
  bdd_done();
  
  return 0; 
}





void  printusage() {
  printf("USAGE:\n");
  printf("       -d <domain file>.nadl   (default \"domain.nadl\")\n");
  printf("       -i <planfile>.plan      (default \"universal.plan\")\n");
  printf("       -o <tracename>.txt      (default \"trace.txt\")\n");
  printf("       -n <bdd node count>     (default 100000)\n");
  printf("       -c <bdd cache size>     (default 10000)\n");
  printf("       -p <part limit>         (default 5000)\n");
  printf("       -v <verbosity level>    (default 0)\n"); 
  printf("       -s system manually                   \n"); 
  printf("          controlled           (default off)\n");
  printf("       -e environment manually              \n"); 
  printf("          controlled           (default off)\n");
  printf("       -h write usage                           \n");
  printf("\n");
}

