/*========================================================================
                              
                                ** UMOP **
		** Universal Multi-Agent OBDD-based Planner **

            Copyright (C) 1999 by Rune M. Jensen and Manuela M. Veloso
            Computer Science Department, Carnegie Mellon University.
            
    Permission is hereby granted to distribute the UMOP software for 
    research purposes, provided this notice is included with any such 
    distribution. 

      *** WE ARE NOT RESPONSIBLE FOR ANY KIND OF DAMAGE TO YOUR  ***
      *** FILES, DATA, HARDWARE, LOSS OF MONEY, SYSTEM CRASHES   *** 
      *** OR ANY OTHER THING YOU MIGHT COME UP WITH.             ***
      *** - USE THIS PROGRAM OF YOUR OWN FREE WILL !!!           ***

========================================================================*/
/****************************************************** 
 * File  : fsm.cc   
 * Desc. : misc. functions for generating the OBDD FSM 
 *         transition relation.
 * Author: Rune M. Jensen
 * Date  : 3/4/99
 ******************************************************/

#include <stream.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <bdd.h>
#include <fstream.h>
#include <set>
#include <string>
#include <vector>
#include <list>
#include "common.hpp"
#include "set.hpp"
#include "dissets.hpp"
#include "fsm.hpp"



/******************************************************
 *              print functions                       *
 ******************************************************/
void agtdef::print() {
  cout << "Actions " << name << "("; 
  if (sclass == sc_env)
    cout << "env) :\n\n";
  else
    cout << "sys) : \n\n";

  for (int i = 0; i < act.size(); i++)
    {
	act[i].print();
	cout << endl;
    }
}


void vardef::print() {
  int i;
  cout << "name:" << name << "\n";
  cout << "type: ";
  switch (type) {
  case vt_bool: 
    cout << "bool\n";
    break;
    case vt_scalar: 
      cout << "scalar\n";
      break;
  }
  cout << "length: " << length << "\nmapping: ";
  for (i=0; i<length; i++)
    cout << i+1 << ":" << var[i] << " ";
}


void domdef::print() {
  cout << "\nDomain Definition\n\n";
  cout << "Agtdef\n\n";
  for (int i = 0; i < agt.size(); i++)
    {
      agt[i].print();
      cout << "\n\n";
    }
  cout << "Vardef:\n\n";
  for (int i = 0; i < var.size(); i++)
    {
      var[i].print();
      cout << "\n\n";
    }

  cout << "\n\nagtnum: " << agt.size();
  cout << "\nvarnames : ";
  for (int i = 0; i < varnames.size(); i++)
    cout << varnames[i] << " ";
  cout << "\nenvnum: " << envnum;
  cout << "\n\n";


}

      

    

/****************************************************
 *  Function building a domain definition           *
 ****************************************************/

//IN  stripped
//  prob: NADL descripton internal rep.
//OUT
// a definition (domdef) of various domain info
// notes: in the BDD variable table current state variables have a similar name as in the definition,
//        next state variables are added a quote.
domdef *mkdomdef(problem *prob) {
  vector<agtdef>     agt;     
  int                actnum,maxacts,length,bvcount;
  int                envnum;
  int               *var;
  vector<vardef>     vard;
  map<string,vardef> vardmap;
  vector<string>     varnames;

  
  char              *str;

  
  //************** build the agtdef  ************** 
  envnum  = 0;
  maxacts = 0;
  actnum  = 0;

  // add environment agents
  for (list<agent>::iterator agti = prob->env.begin(); agti != prob->env.end(); ++agti)
    {
      vector<action> actv;
      list<action>::iterator acti;
      for (acti = agti->actions.begin(); acti != agti->actions.end(); ++acti)
	actv.push_back(*acti);

      agt.push_back(agtdef(agti->name,sc_env,actv));
      
      actnum += actv.size();
      if (actv.size() > maxacts) maxacts = actv.size();
    }


  envnum = agt.size();
  // add system agents
  for (list<agent>::iterator agti = prob->sys.begin(); agti != prob->sys.end(); ++agti)
    {
      vector<action> actv;
      for (list<action>::iterator acti = agti->actions.begin(); acti != agti->actions.end(); ++acti)
	actv.push_back(*acti);

      agt.push_back(agtdef(agti->name,sc_sys,actv));
      
      if (actv.size() > maxacts) maxacts = actv.size();
    }

  // post: maxacts: the maximal number of actions


  //************** build the vardef ******************
  // Interleaved Ordering of S E and S' E':
  //
  // A1,A2,...,Am,S11,S11',...,S1nS'1n, S21,S21',... ,S2m,S2m' ...
  //
  //**************************************************

  bvcount = 0;   // first bdd variable number

  // start with the action identifier variables
  for (int i=0;i < agt.size();i++)
    {
      if (agt[i].act.size() == 1)
	length = 1; // redundant, but we want to keep all information in the BDD
      else
	length = int(ceil(log(agt[i].act.size())/log(2)));

      // make var to BDDvar  mapping
      var = new int[length];
      for (int j=0;j<length; j++)
	var[j] = bvcount + j;
	
      // save var description in vector and map
      str = new char[MAXNAMELENGTH];
      sprintf(str,"%d",i);
      vard.push_back(vardef(str,vt_scalar,length,var));
      vardmap[str] = vardef(str,vt_scalar,length,var);
      bvcount += length;
    }


  // continue with the variables S11,S11',...,S1nS'1n, S21,S21',... ,S2m,S2m' 
  // (and save the var names in a list)
  for (list<vardecl>::iterator vardi = prob->vars.begin(); vardi != prob->vars.end(); ++vardi)
    {
      // find the bdd variable length of variable type 
      if (vardi->type->type == vt_bool) 
	length = 1;
      else
	length = vardi->type->range;
      
      // assign bdd variable space for each variable
      for (set<string>::iterator si = vardi->ids.begin(); si != vardi->ids.end(); ++si)      
	{
	  // S
	  
	  // make mapping to bddvars
	  var = new int[length];
	  for (int i = 0; i<length; i++)
	    var[i] = bvcount + 2*i;
	  
	  // save var description in vector and map
	  vard.push_back(vardef(*si,vardi->type->type,length,var));
	  vardmap[*si] = vardef(*si,vardi->type->type,length,var);
	  
	  // add variable to variable list
          varnames.push_back(*si);

	  // S'
	  // make mapping to bddvars
	  var = new int[length];
	  for (int i = 0; i<length; i++)
	    var[i] = bvcount + 2*i + 1;
	  
	  // save var description in vector and map
	  vard.push_back(vardef(*si+"'",vardi->type->type,length,var));	  
	  vardmap[*si+"'"] = vardef(*si+"'",vardi->type->type,length,var);
	  
	  bvcount += 2*length;
	}      
    }

  return new domdef(agt,envnum,actnum,vard,vardmap,varnames,bvcount);  
}


/******************************************************
 * Dynamic BDD var autoreorder definition             *
 ******************************************************/
// IN
//  def : domain definition
// OUT (side effect)
//  A BDD block definition for each action and variable.
//  action varibles and current and next state varibles
//  are assumed to be defined on a continuous block of 
//  variables 
void mkbddblocks(domdef *def) {
  int    i,j,start,end;
  

  // clear some old block definition
  bdd_clrvarblocks();

  // add blocks for action variables
  for (i=0; i<def->agt.size(); i++)
    {

      // find the start and end of the action block
      start = def->bddvarnum; 
      end = 0;
      for (j=0; j < def->var[i].length; j++)
	{
	  if (def->var[i].var[j] < start) start = def->var[i].var[j];
	  if (def->var[i].var[j] > end)   end = def->var[i].var[j];
	}
      
      // add variable block for dynamic bdd var reordering
      if (bdd_intaddvarblock(start,end,BDD_REORDER_FIXED) < 0)
	{
	  cout << "fsm.cc mkbddblocks : cannot make var. block ";
	  cout << start << "-" << end << "\nexiting\n\n";
	  exit(1);
	}
    }


  // add blocks for variables
  for (i = 0; i < def->varnames.size(); i++)
    {
      string id = def->varnames[i];
      vardef s(def->varmap[id]);
      vardef sm(def->varmap[id + "'"]);
      
      // find the start and end of the variable block
      start = def->bddvarnum;
      end = 0;
      for (j=0; j < s.length; j++)
	{
	  if (s.var[j] < start) start = s.var[j];
	  if (s.var[j] > end)   end = s.var[j];
	}

      for (j=0; j < sm.length; j++)
	{
	  if (sm.var[j] < start) start = sm.var[j];
	  if (sm.var[j] > end)   end = sm.var[j];
	}
      
      // add variable block for dynamic bdd var reordering
      if (bdd_intaddvarblock(start,end,BDD_REORDER_FIXED) < 0)
	{
	  cout << "fsm.cc mkbddblocks : cannot make var. block ";
	  cout << start << "-" << end << "\nexiting\n\n";
	  exit(1);
	}
    }
}  



/******************************************************
 * Function defining bdd variable loactions           *
 ******************************************************/

// IN
//  def  : domain definition. (includes description
//         of bdd variable locations.
// OUT   
//  instantiation of BuDDy variable location desriptions
//  and variable relations.
varrels *mkvarrels(domdef *def) {
  
  int     i,j,k,p;
  varrels *rels;

  rels = new varrels;

  rels->old2new = bdd_newpair();
  rels->new2old = bdd_newpair();
  
  // relate new/old and old/new state variables
  for (j = 0; j < def->varnames.size(); j++)
    {
      string id = def->varnames[j];
      vardef oldv(def->varmap[id]);
      vardef newv(def->varmap[id + "'"]);
      for (i=0; i < newv.length; i++)
	{
	    bdd_setpair(rels->old2new,oldv.var[i],newv.var[i]);
	    bdd_setpair(rels->new2old,newv.var[i],oldv.var[i]);
	}
    }

  // intarray representing a set of BDD variables (when calling bdd_makeset
  // the size is specified, so this var can be shared for the 4 calculation below) 
  int* varset = new int[def->bddvarnum];


  //************** full state variable definitions ****************
  
  // define the full set of old state variables
  j = 0;
  for (k = 0; k < def->varnames.size(); k++)
    {
      vardef oldv(def->varmap[def->varnames[k]]);
      for (i=0; i < oldv.length; i++)
	varset[j++] = oldv.var[i];
    }
  rels->oldstatefull = bdd_makeset(varset,j);


  // define the full set of new state variables
  j = 0;
  for (k = 0; k < def->varnames.size(); k++)
    {
      vardef newv(def->varmap[def->varnames[k] + "'"]);
      for (i=0; i < newv.length; i++)
	varset[j++] = newv.var[i];
    }
  rels->newstatefull = bdd_makeset(varset,j);


  // define environment action variables
  j = 0;
  for (i=0; i<def->envnum; i++)
    {
      char str[MAXNAMELENGTH];
      sprintf(str,"%d",i);
      vardef actv(def->varmap[str]);
      for (k=0; k < actv.length; k++)
	varset[j++] = actv.var[k];
    }
  rels->envactions = bdd_makeset(varset,j);


  // define system action variables
  j = 0;
  for (i=def->envnum; i<def->agt.size(); i++)
    {
      char str[MAXNAMELENGTH];
      sprintf(str,"%d",i);
      vardef actv(def->varmap[str]);
      for (k=0; k < actv.length; k++)
	varset[j++] = actv.var[k];
    }
  rels->sysactions = bdd_makeset(varset,j);


return rels;
}
  


/******************************************************
 * Function adjusting domdef to a variable ordering   *
 ******************************************************/

// IN
//  order : Array of variable levels in ordering
//  def   : domain definition
// OUT   
//  sideeffect: adjusted domdef
void adjust2order(int *ordering, domdef *def) {
  
  // go through all domain variables and map the
  // variable locations to the new ordering
  for (int j=0; j < def->var.size(); j++)
      for (int i=0; i < def->var[j].length; i++)
	def->var[j].var[i] = ordering[def->var[j].var[i]];
}





/******************************************************
 * Functions for extracting current state variables   *
 * used in an action expression (output is a list of  *
 * variables)                                         *                   
 ******************************************************/


/******************************************************
 * Functions for generating OBDD's from formulas      *
 ******************************************************/

bdd formula2bdd(formula *f, domdef *def) {

bdd b,l,r,e;

  switch (f->type) {
  case ft_atom:
    b = atom2bdd(f->atomic, def);
    break;

  case ft_neg:
    b = formula2bdd(f->f1, def);
    b = !b;
    break;
    
  case ft_and:
    l = formula2bdd(f->f1,def);
    r = formula2bdd(f->f2,def);
    b = l & r;
    break;
    
  case ft_or:
    l = formula2bdd(f->f1,def);
    r = formula2bdd(f->f2,def);
    b = l | r;
    break;
    
  case ft_impl:
    l = formula2bdd(f->f1,def);
    r = formula2bdd(f->f2,def);
    b = l >> r;
    break;
    
  case ft_biimpl:
    l = formula2bdd(f->f1,def);
    r = formula2bdd(f->f2,def);
    b = (l >> r) & (l << r);
    break;
    
  case ft_ite:
    e = formula2bdd(f->f1,def);
    l = formula2bdd(f->f2,def);
    r = formula2bdd(f->f3,def);
    b = (e & l) | (!e & r); 
    break;

  case ft_true:
    b = bddtrue;
    break;

  case ft_false:
    b = bddfalse;
    break;

  case ft_paren:
    b = formula2bdd(f->f1,def);
    break;
  }
return b;
}



bdd atom2bdd(atom *a, domdef *def) {
  
  bdd b;
  
  switch (a->type) {
    case at_numprop:
      a->prop->minus2plus();
      b = numprop2bdd(a->prop,def);
      break;
    case at_boolvar:
      vardef vd(def->varmap[a->var]);
      if (vd.var != NULL) 
	b = bdd_ithvar(vd.var[0]);
      else
	{
	  cout << "\nfsm.hpp, atom2bdd: variable " << a->var;
	  cout << " is not defined\nexiting\n";
	  exit(1);
	}
      break;
  }
  return b;
}
 

//** Functions for generating BDD's from numprops  ***

bdd numprop2bdd(numberprop *prop,domdef *def) {
  bdd b,e,ld,rd;
  binary *l, *r;
  numberexp *c;
  numberprop *propc;
  int i,j,digitnum;

  switch (prop->op) {
  case ro_lt:
    c = prop->left;
    prop->left = prop->right;
    prop->right = c;
    prop->op = ro_gt;
    b = numprop2bdd(prop,def); 
    break;
  case ro_ne:
    propc = new numberprop;
    propc->op = ro_eq;
    propc->left  = prop->left;
    propc->right = prop->right;
    b = !numprop2bdd(propc,def);
    break;
  case ro_eq:
    l = numexp2binary(prop->left,def);
    r = numexp2binary(prop->right,def);
    
    b = bddtrue;
    
    // test if l = r
    i = 0;
    while (i < l->digitnum || i < r->digitnum)
      {
	if (i < l->digitnum && i < r->digitnum)
	  {
	    b &= l->digit[i] >> r->digit[i];
	    b &= l->digit[i] << r->digit[i];
	  }
	else
	  if ( i < l->digitnum )
	    {
	      // r->digit[i] = bddfalse
	      b &= !l->digit[i];
	    }
	  else
	    {
	      // l->digit[i] = bddfalse
	      b &= !r->digit[i];
	    }
	i++;
      }
    break;

  case ro_gt:
    l = numexp2binary(prop->left,def);
    r = numexp2binary(prop->right,def);
    
    digitnum = l->digitnum > r->digitnum ? l->digitnum : r->digitnum;
    
    b = bddfalse;
    for (i=0; i<digitnum; i++)
      {
	// all digits before digit_i must be equal
	e = bddtrue;
	for (j = i+1; j<digitnum; j++)
	  {
	    ld = j >= l->digitnum ? bddfalse : l->digit[j];
	    rd = j >= r->digitnum ? bddfalse : r->digit[j];
	    e &= ld >> rd;
	    e &= ld << rd;
	  }

	ld = i >= l->digitnum ? bddfalse : l->digit[i];
	rd = i >= r->digitnum ? bddfalse : r->digit[i];

	// if e is true then if ld > rd: l > r
	b |= e & ld & !rd; 
      }
    break;
  }
  return b;
}


binary *numexp2binary(numberexp *num,domdef *def) {
  binary *b,*l,*r;
  bdd    *digit,c;
  int     digitnum,i,number;

  switch (num->type) {
  case nt_minus:
    cout << "\n\nfsm.cc numexp2binary: error type of numexp cannot be nt_minus\nexiting\n";
    exit(1);
    break;

  case nt_number:
    number = num->number;
    // the digitnum of 0 is 0, of 1 is 1, of 2,3 is 2, of 4,5,6,7 is 3 etc.
    digitnum = int(ceil(log(number+1)/log(2)));
    
    if (digitnum > 0)
      digit = new bdd[digitnum];
    else
      digit = NULL;
    
    // initialize digit array
    for (i=0; i<digitnum; i++)
      {
	if (number % 2)
	  digit[i] = bddtrue;
	else
	  digit[i] = bddfalse;
	number /= 2;
      }
    
    b = new binary(digitnum,digit);
    break;

  case nt_var:
    {
      vardef vard(def->varmap[num->var]);
      if (vard.var == NULL) 
	{
	  cout << "\nfsm.hpp numexp2binary: variable " << num->var;
	  cout << " is not defined\nexiting\n";
	  exit(1);
	}      
      else
	{
	  digitnum = vard.length;
	  
	digit = new bdd[digitnum];
	
	//intialize the digit array
	for (i=0; i < digitnum; i++)
	  digit[i] = bdd_ithvar(vard.var[digitnum - 1 - i]);
	
	b = new binary(digitnum,digit);
	}
    }
    break;

  case nt_plus:
    l = numexp2binary(num->left,def);
    r = numexp2binary(num->right,def);

    // define result binary
    digitnum = l->digitnum > r->digitnum ? l->digitnum + 1 : r->digitnum + 1;
    digit = new bdd[digitnum];

    // set initial carry
    c = bddfalse;
    i = 0;

    // add the binaries
    while ( i < l->digitnum || i < r->digitnum )
      {
	if ( i < l->digitnum && i < r->digitnum )
	  {
	    digit[i] = l->digit[i] ^ r->digit[i] ^ c;
	    c = l->digit[i] & r->digit[i] |
	        l->digit[i] & c |
	        r->digit[i] & c;
	  }
	else
	  if ( i < l->digitnum )
	    {
	      //r->digit[i] = bddfalse
	      digit[i] = l->digit[i] ^ c;
	      c = l->digit[i] & c;
	    }
	  else
	    {
	      //l->digit[i] = bddfalse
	      digit[i] = r->digit[i] ^ c;
	      c = r->digit[i] & c;
	    }
	i++;
      }
    
    // post: i = last digit in result (corresponds to carry)
    digit[i] = c;

    b = new binary(digitnum,digit);
    break;
  }
  return b;
}




/******************************************************
 * Functions for Generating the subrelations A, F and I
 *
 ******************************************************/

// IN
//   var    : bdd var mapping
//   length : number of bdd vars
//   n      : integer to encode.
//   OUT
//   bdd : BDD with the binary encoding of n
//         starting with the most significant bit. 
bdd int2bdd(int *var,int length,int n)  {
  bdd res;
  int j;  

  res = bddtrue;
  for (j=length-1; j >= 0; j--)
    {
      if (n % 2) 
	res &= bdd_ithvar(var[j]); 
      else
	res &= bdd_nithvar(var[j]);
      n /= 2;
    }
return res;
}




// I(i,j,def)
// IN
//  i,j : Action j of agent i.  
//  def : domain definition. 
//
// OUT
//  the bdd identifier expression
//  for action(i,j).
bdd I(int i, int j, domdef *def) {

  char str[MAXNAMELENGTH];
  sprintf(str,"%d",i);
  vardef vd(def->varmap[str]);
  return int2bdd(vd.var,vd.length,j);
}  



// F(v,i,j,def)
// IN
//  v        : variable to check
//  i,j      : Action j of agent i.  
//  def      : Domain definition 
// 
// OUT
//  bddtrue : var is unconstrained by action(i,j).
//  bddfalse: var is constrained by action(i,j)
//            (var is in the action's curvar or con: list.)
bdd F(string v, int i, int j, domdef *def) {

  if (setMember(def->agt[i].act[j].con,v))
    return I(i,j,def) >> bddfalse; // v is not free in Act 
  else
    return I(i,j,def) >> bddtrue;  // v is free

} 



//***** Functions for constraining the action tuple *********

// ADEF(def)
// IN
//  def : domain definition. 
// OUT
//  bdd constraining possible actions to only defined actions
bdd ADEF(int agt,domdef *def) {
  bdd res;
  int i;
  formula *f;
  atom *a;
  numberprop *p;
  numberexp  *l,*r;
  char str[MAXNAMELENGTH];

  f = new formula;
  a = new atom;
  p = new numberprop;
  l = new numberexp;
  r = new numberexp;
  
  // define formula : "actid" < max_val+1
  f->type = ft_atom;
  f->atomic = a;
  a->type = at_numprop;
  a->prop = p;
  p->op = ro_lt;
  p->left = l;
  p->right = r;
  l->type = nt_var;
  r->type = nt_number;
  
  // constrain action tuple to only defined actions
  res = bddtrue;
  sprintf(str,"%d",agt);
  l->var = str;
  r->number = def->agt[agt].act.size();
  res &= formula2bdd(f,def);

  return res;
}




// IND(i1,j1,i2,j2,def)
// IN
//  i1,j1 : action identifier 1
//  i2,j2 : action identifier 2
//  def : domain definition. 
// OUT
//  bool expressing that A1 
//  is consistent with A2
//  (no overlap of constrained variables)
int IND(int i1, int j1, int i2, int j2, domdef *def) {

  set<string> con1,con2;
  
  con1 = def->agt[i1].act[j1].con;
  
  for (set<string>::iterator id1 = con1.begin(); id1 != con1.end(); ++id1)
    {
      con2 = def->agt[i2].act[j2].con;
      for (set<string>::iterator id2 = con2.begin(); id2 != con2.end(); ++id2)
	  if (*id1 == *id2) return 0;
    }
  return 1;
}


// AC(def)  (the I subrelation in thesis)
// IN
//  def : domain definition. 
// OUT
//  bdd expressing the total independency and definition constraint 
//  on action tuples
bdd AC(domdef *def) {
  bdd res,l,r;
  int i1, i2, j1, j2,i;


  // constrain action tuple to posible values
  res = bddtrue;
  for (i=0; i<def->agt.size(); i++)
    res &= ADEF(i,def);
 
  // concurrent independency constraints
  
  // add internal independency condition
  // environment agents  
  // go through all possible non-symmetric action pairs
  for (i1=0; i1 < def->envnum - 1; i1++)
    for (i2=i1+1; i2 < def->envnum; i2++)
      for (j1=0; j1 < def->agt[i1].act.size(); j1++)
  	for (j2=0; j2 < def->agt[i2].act.size(); j2++)
	  if (!IND(i1,j1,i2,j2,def))
	    res &= ! ( I(i1,j1,def) & I(i2,j2,def) );
	
  // system agents 
  // go through all possible non-symmetric action pairs
  for (i1=def->envnum; i1 < def->agt.size() - 1; i1++)
    for (i2=i1+1; i2 < def->agt.size(); i2++)
      for (j1=0; j1 < def->agt[i1].act.size(); j1++)
  	for (j2=0; j2 < def->agt[i2].act.size(); j2++)
	  if (!IND(i1,j1,i2,j2,def))
	    res &= ! ( I(i1,j1,def) & I(i2,j2,def) );
	
  return res;
}



  
// P(i,j,def)
// IN
//  i,j : Action j of agent i.  
//  def : domain definition. 
//
// OUT
//  bdd expression for the precondition
//  of action(i,j).
bdd P(int i, int j, domdef *def) {
  return formula2bdd(def->agt[i].act[j].pre,def);
}

// E(i,j,def)
// IN
//  i,j : Action j of agent i.  
//  def : domain definition. 
//
// OUT
//  bdd expression for the effect
//  of action(i,j).
bdd E(int i, int j, domdef *def) {
  return formula2bdd(def->agt[i].act[j].eff,def);
}


// IV(partno,def) (the F subrelation in thesis(but only for one cluster))
// IN
//  def      : domain definition. 
//  var      : variables constrained by cluster
//
// OUT
//  bdd expression for the invariable expression
//  constraining only new vars in the partition
// IV(i) = /\v in part(i) ((/\i,j F(i,j,v)) => v' = v)
bdd IV(set<string> vars,domdef *def) {
  int i,j;
  bdd    l,r,res;

  res = bddtrue;
  
  for (set<string>::iterator var = vars.begin(); var != vars.end(); ++var)
    { 

      // make left side
      l = bddtrue;
      for (i=0; i < def->agt.size(); i++)
	for (j=0; j < def->agt[i].act.size(); j++)
	  l &= F(*var,i,j,def);
      
      // make right side
      vardef s(def->varmap[*var]);
      vardef sm(def->varmap[*var + "'"]);
      r = bddtrue;
      for (i=0; i < s.length; i++)
	{
	  r &= bdd_ithvar(s.var[i]) >> bdd_ithvar(sm.var[i]);
	  r &= bdd_ithvar(s.var[i]) << bdd_ithvar(sm.var[i]);
	}
      
      // add to result
      res &= l >> r;
      
      if (verbosity > 1)
	cout << "Finished var: " << *var << "\n";
    }
return res;
}
	

// A(i,j,def) (the A subrelation in thesis (but only for one action))
// IN 
//  i : agentt number
//  j : action number
//  def : domain definition
// OUT
//  bdd representing the action
//  A =  I(i,j) => P(i,j) /\ E(i,j) 
bdd A(int i,int j,domdef *def) {
  bdd l,r;

  l = bddtrue;
  r = bddtrue;

  // make left side of impl.
  l &= I(i,j,def);

  // add relation constraint
  r &= P(i,j,def) & E(i,j,def);
    
  return l >> r;
}




/****************************************************** 
 * Functions for building conjunctive partitioned
 * transitionsrelations
 *
 * Current simplification: 
 * - environment actions are represented in the transi-
 *   tion relation to simplify research on Game Universal
 *   plans
 *
 * Thus:
 * T(S,As,Ae,S') = T1(S,As,Ae,S') /\ ... /\ T2(S,As,Ae,S')
 ******************************************************/



void trel::print() {
  int i;

  if (this) 
    {
      cout << "Number of partitions : " << partnum << "\n";
      for (i=0; i<partnum; i++)
	{
	  cout << "  Partition # : " << i << "\n";
	  p[i].print();
	}
    }
}

void relpartition::print() {
  cout << "   Exp nodecount : " << bdd_nodecount(exp) << "\n";
  cout << "   BDD varset : "; bdd_printset(varset); cout << "\n";
  cout << "   BDD varset (nodecount) : " << bdd_nodecount(varset) << "\n";
}



char *stringof(char *buf,string s) {
  return strcpy(buf,s.c_str());
}

// mkpartitions(def) 
// IN
//  def  : domain definition. 
// OUT
//  A maximal conjunctive paritioning: T(s,as,s') = T1(s,as,s') /\ T2(s,as,s') ...
//  not depending on environment actions
trel *mkpartitions(domdef *def,bool adversarial) {    

  int i,j,k,p,partnum;
  char buf[MAXNAMELENGTH];
  relpartition *part;

  // clustering variables
  disset *dset; 
  disnode *set1,*set2;


  //************** find clusters of nextstate vars  ************
  
  // make initial disjoint set
  // (a singleton set for each variable)
  dset = new disset();
  for (k = 0; k < def->varnames.size(); k++)
    dset->makeset(def->varnames[k]);
   
  // find disjoint set for next state variable clusters
  for (i=0; i < def->agt.size(); i++)
    for (j=0; j < def->agt[i].act.size(); j++)
      for (set<string>::iterator ilp = def->agt[i].act[j].con.begin(); ilp != def->agt[i].act[j].con.end(); ++ilp) 
	if (ilp == def->agt[i].act[j].con.begin())
	  set1 = dset->find(*ilp);
	else
	  {
	    // union subsets of each variable in the variable list
	    set2 = dset->find(*ilp);
	    if (set1 != set2)
	      set1 = dset->setunion(set1,set2);
	  }


  // for each agent, find the set of next state variables
  // constrained by some action of the agent.
  vector< set<string> >  agtactvar(def->agt.size()); 
  for (i=0; i < def->agt.size(); i++)
    for (j=0; j < def->agt[i].act.size(); j++)
      agtactvar[i] = setUnion(def->agt[i].act[j].con,agtactvar[i]);

 
  // union overlapping sets of vars constrained by an env. agent
  // obs this only effects partitions constraining next state variables constrained
  // by the environment. 

  for (i=0; i < def->envnum; i++)
    for (set<string>::iterator ilp = agtactvar[i].begin(); ilp != agtactvar[i].end(); ++ilp)       
      if (ilp == agtactvar[i].begin())
	  set1 = dset->find(*ilp);
      else
	{
	  // union subsets of each variable in the variable set
	  set2 = dset->find(*ilp);
	  if (set1 != set2)
	    set1 = dset->setunion(set1,set2);
	}
  
    
  // make variable table for clusters
  dset->mkdisjointsets();

  if (verbosity > 0) cout << "Number of disjoint next state and env. agent variable sets : " << dset->subsetnum << "\n";

  // debug write which partitions each environment agent belongs to 
  if (verbosity > 1)
    for (i = 0; i < dset->subsetnum; i++)
      {
	cout << "Env. agents in partition " << i << " ("; setPrint(dset->disjointset[i]); cout << ") : "; 
	for (j = 0; j < def->envnum; j++)
	  if (!setIntersection(agtactvar[j],dset->disjointset[i]).empty())
	    {
	      cout << def->agt[j].name << "("; setPrint(agtactvar[j]); cout << ") ";
	    }
	
	cout << endl;
      }
	
    

  //************** make partitions ************


  // number of partitions (obs this might be reduced during Ranjan optimization)
  partnum = dset->subsetnum; 
  part = new relpartition[partnum]; 

  // ************************************************************************
  // Add partition expressions
  // ************************************************************************
  // insert IV in partitions
  for (i=0; i<partnum; i++)
      part[i].exp = IV(dset->disjointset[i],def);

  // insert A in partitions
  for (i=0; i < def->agt.size(); i++)
    for (j=0; j< def->agt[i].act.size(); j++)
      {
	// find the cluster number of the action
	// lookup the first variable or else set cluster number to 0
	if ( def->agt[i].act[j].con.size() > 0 )
	  p = dset->find(*(def->agt[i].act[j].con.begin()))->setnum;
	else
	  p = 0;

	part[p].exp &= A(i,j,def);

	if (verbosity > 1)
	  cout << "Finished action(" << i << "," << j << ")\n";
      }

  
  // insert action var value constrain

  for (i=0; i<def->agt.size(); i++)
    if (agtactvar[i].size() > 0) // if any vars constrained by agent
	{
	  p = dset->find(*agtactvar[i].begin())->setnum;// find the cluster of the agent
	  part[p].exp &= ADEF(i,def); // constrain values of agt act var.
	}


  // Insert the I relation to the partitions


  int i1, i2, j1, j2;
 
  // concurrent independency constraints
  
  // add internal independency condition
  // environment agents  
  // go through all possible non-symmetric action pairs
  for (i1=0; i1 < def->envnum - 1; i1++)
    for (i2=i1+1; i2 < def->envnum; i2++)
      for (j1=0; j1 < def->agt[i1].act.size(); j1++)
  	for (j2=0; j2 < def->agt[i2].act.size(); j2++)
	  if (!IND(i1,j1,i2,j2,def))
	    {
	      // find the cluster of the constraint by looking at the first agent
	      p = dset->find(*agtactvar[i1].begin())->setnum;
	      part[p].exp &= !(I(i1,j1,def) & I(i2,j2,def));
	    }
  
  // system agents 
  // go through all possible non-symmetric action pairs
  for (i1=def->envnum; i1 < def->agt.size() - 1; i1++)
    for (i2=i1+1; i2 < def->agt.size(); i2++)
      for (j1=0; j1 < def->agt[i1].act.size(); j1++)
  	for (j2=0; j2 < def->agt[i2].act.size(); j2++)
	  if (!IND(i1,j1,i2,j2,def))
	    {
	      // find the cluster of the constrain by looking at the first agent
	      p = dset->find(*agtactvar[i1].begin())->setnum;
	      part[p].exp &= !(I(i1,j1,def) & I(i2,j2,def));
	    }
  

  // ************************************************************************
  // Add next state variable sets 
  // ************************************************************************
  int *varset = new int[def->bddvarnum];

  for (p=0; p<partnum; p++)
    { 
      j = 0;
      for (set<string>::iterator ilp = dset->disjointset[p].begin(); ilp != dset->disjointset[p].end(); ++ilp) 
	{
	  vardef vd(def->varmap[*ilp+"'"]);
	  for (i=0 ; i < vd.length; i++)
	    varset[j++] = vd.var[i];
	}
      part[p].varset = bdd_makeset(varset,j);
    }


  // ************************************************************************
  // Existentially quantify env actions in partition if
  // algorithm is non-adversarial
  // ************************************************************************


  if (!adversarial)
    for (p=0; p<partnum; p++)
      part[p].exp = bdd_exist(part[p].exp, def->rels->envactions);
 

return new trel(partnum,part);
}






/****************************************************** 
 * Ranjan optimization of partitions 
 * 
 ******************************************************/


// partcmp(p1,p2)
// IN
//  p1,p2 : two partition variable sets to compare
// OUT
//  - if p1.varset < p2.varset
//  0 if p1.varset = p2.varset
//  + if p1.varset > p2.varset
int partcmp(relpartition *p1, relpartition *p2) {
  if (bdd_nodecount(p1->varset) < bdd_nodecount(p2->varset))
    return 1;
  else if (bdd_nodecount(p1->varset) == bdd_nodecount(p2->varset))
    return 0;
  else 
    return -1;
}



// IN
//  T : A transition relation partition relation set. 
// OUT
//  side effect: A sorted transition relation partition set
//  (the partition quantifying most variables is ordered first)
void sorttrel(trel *T) {
  qsort(T->p, T->partnum, sizeof(relpartition),(int (*)(const void *,const void *)) partcmp);
}




		
//IN
// T         : Partitioned transition relation
// threshold : least number of nodes in optimized transition relation
//OUT
// transition relation where partitions are unified until partition size
// is above threshold and sorted s.g. partitions with the greatest
// number of variables are quantified first		       
void ranjanoptimize(trel *T,int threshold) {	      						     
  
  int part = 0;
  int i = 0;
  int add;

  // collapse partitiones s.t. the node count of each is above the threshold
  while (i<T->partnum)
    {
      T->p[part].exp    = T->p[i].exp;
      T->p[part].varset = T->p[i].varset;
      i++;
      add = 1;
      while (add)
	{
	  if (i<T->partnum)
	    if (bdd_nodecount(T->p[part].exp & T->p[i].exp) < threshold )	    
	      {
		T->p[part].exp    &= T->p[i].exp;
		T->p[part].varset &= T->p[i].varset;
		i++;
	      }
	    else
	      add = 0;
	  else
	    add = 0;
	}
      part++;
    }
  T->partnum = part;
  
  sorttrel(T); 

  if (verbosity > 0)
    cout << "Number of Ranjan optimized partitions : " << part << "\n";

}	  
  




