/*========================================================================
                              
                                ** UMOP **
		** Universal Multi-Agent OBDD-based Planner **

            Copyright (C) 1999 by Rune M. Jensen and Manuela M. Veloso
            Computer Science Department, Carnegie Mellon University.
            
    Permission is hereby granted to distribute the UMOP software for 
    research purposes, provided this notice is included with any such 
    distribution. 

      *** WE ARE NOT RESPONSIBLE FOR ANY KIND OF DAMAGE TO YOUR  ***
      *** FILES, DATA, HARDWARE, LOSS OF MONEY, SYSTEM CRASHES   *** 
      *** OR ANY OTHER THING YOU MIGHT COME UP WITH.             ***
      *** - USE THIS PROGRAM OF YOUR OWN FREE WILL !!!           ***

========================================================================*/
/****************************************************** 
 * File  : plan.cc  
 * Desc. : Strong StrongCyclic and Optimistic  planning 
 *         algorithms
 *         ver. 1.1. : The SCA and OA planning algorithm
 * Author: Rune M. Jensen CS, CMU
 * Date  : 4/4/99, 5/25/00, 10/26/00, 7/16/01
 ******************************************************/

#include <bdd.h>
#include "common.hpp"
#include "domain.hpp"
#include "fsm.hpp"
#include "plan.hpp"
#include "bddprint.hpp"



/****************************************************** 
 *      Aux. functions   
 *     
 ******************************************************/


bdd prunestates(bdd preimage, bdd acc)  {
  return preimage & !acc;
}


bdd projact(bdd sa,domdef *def)  {
  return bdd_exist(sa, def->rels->sysactions);
}


// IN
//  tr   : Partitioned transition relation
//  acc  : currently visited states (encoded in current state variables)
//  rels : bdd variable info structure 
// OUT
//  strongpreimage of acc encoded in act * current state vars
//  the calculation combines a partitioned representation 
bdd strongpreimage(trel *tr, bdd acc,domdef *def)  {
  bdd p1,p2,c;
  int i;

  c = bdd_replace(acc,def->rels->old2new);

  p1 = c;
  // P1 preimage composition 
  for (i=0; i < tr->partnum; i++)
    {
      p1 = bdd_appex(tr->p[i].exp,p1,bddop_and,tr->p[i].varset);
      if (verbosity > 1)
	cout << "Finished Strong partition A" << i << " node count: " 
	     << bdd_nodecount(p1) << "\n";
    }

  p2 = !c;
  // P2 preimage composition 
  for (i=0; i< tr->partnum; i++)
    {
      p2 = bdd_appex(tr->p[i].exp,p2,bddop_and,tr->p[i].varset);
      if (verbosity > 1)
	cout << "Finished Strong partition B" << i << " node count: " 
	     << bdd_nodecount(p2) << "\n";
    }

  return p1 & !p2;
}


// IN
//  tr   : Partitioned transition relation
//  acc  : currently visited states (encoded in current state variables)
//  rels : bdd variable info structure 
// OUT
//  weak preimage of acc encoded in act * current state vars
//  the calculation combines a partitioned representation 
bdd weakpreimage(trel *tr, bdd acc,domdef *def)  {
  bdd res;
  int i;

  res = bdd_replace(acc,def->rels->old2new); 
  // preimage composition 
  for (i=0; i<tr->partnum; i++)
    {
      res = bdd_appex(tr->p[i].exp,res,bddop_and,tr->p[i].varset);
      if (verbosity > 1)
	cout << "Finished Weak partition " << i << " node count: " 
	     << bdd_nodecount(res) << "\n";
    }
  return res;
}





/****************************************************** 
 *          SP  
 *          Strong Planning
 ******************************************************/
// IN
//  tr     : Partitioned transition relation
//  init   : Initial state (encoded in current state vars)
//  goal   : Goal state (encoded in current state vars)
//  sa     : plan output handle
//  def    : Domain definition (containing a description of the 
//           location of variables in bdds.
// OUT
//  1:  some plan found
//  0:  fix point found, but Init is not a subset 
int SP(trel *tr, bdd init, bdd goal, bdd *sa,domdef *def)
{
  bdd acc = goal;                // acc encoded using new state vars (visited states)
  bdd preimage, prunedpreimage; 
  int i = 0;
  
  
  //main loop
  while ((acc & init) != init)
    {
      preimage = strongpreimage(tr,acc,def); // act*current bdd
      prunedpreimage = prunestates(preimage,acc); // act*current bdd

      
      if (prunedpreimage != bddfalse)
	{
	  *sa |= prunedpreimage;
	  acc |= projact(prunedpreimage,def);
	  if (verbosity > 0)
	    cout << "Added preimage " << i << "\n";
	  i++;
	}
      else
	// fix point found, but init not included
	return 0;
    }
  // init is included
  return 1;
}



/****************************************************** 
 *          SCP  
 *          Strong Cyclic Planning
 ******************************************************/


bdd OSA(trel *tr,bdd CSA,bdd s,domdef *def) {
  bdd res,c;
  int i,partnum;
  varrels *rels;

  partnum = tr->partnum;
  rels = def->rels;

  c = bdd_replace(s,rels->old2new);

  res = !c;

  // OSA composition 
  for (i=0; i<partnum; i++)
    {
      res = bdd_appex(tr->p[i].exp,res,bddop_and,tr->p[i].varset);
      if (verbosity > 1)
	cout << "OSA Finished partition " << i << " node count: " 
	     << bdd_nodecount(res) << "\n";
    }

  return res & CSA;
}
  



// IN
//  tr     : Partitioned Transition relation 
//  acc    : Visited states from goal
//  sa     : current s->a table
//  def    : Domain definition (containing a description of the 
//           location of variables in bdds.)
// OUT
//  update of SA and acc
void faircycles(trel *tr, bdd *acc, bdd *sa,bdd *oldacc, domdef *def) {
  varrels *rels;
  bdd IWpi, Outgoing;

  bdd WpiAcc = bddfalse;
  bdd OldWpiAcc = bddtrue;
  bdd CSA = bddfalse;

  rels = def->rels;

  int i = 0;

  while (CSA == bddfalse && OldWpiAcc != WpiAcc)
    {
      IWpi = weakpreimage(tr,projact(WpiAcc,def) | *acc,def);
      CSA = prunestates(IWpi,*acc);
      Outgoing = CSA;
      while (CSA != bddfalse && Outgoing != bddfalse)
	{
	  Outgoing = OSA(tr,CSA,projact(CSA,def) | *acc,def); 
	  CSA &= !Outgoing;
	}
      if (CSA != bddfalse)
	{
	  *sa |= CSA;
	  *oldacc = *acc;
	  *acc |= projact(CSA,def);
	  return;
	}
      else
	{
	  OldWpiAcc = WpiAcc;
	  WpiAcc |= IWpi;
	  if (verbosity > 0)
	    cout << "   Fair layer " << i++ << endl;
	} 
    }
  *oldacc = *acc;
}






// IN
//  tr     : Partitioned transition relation
//  init   : Initial state (encoded in current state vars)
//  goal   : Goal state    (encoded in current state vars)
//  sa     : plan output handle
//  def    : Domain definition 
// OUT
//  1:  some plan found
//  0:  no plan found 
int SCP(trel *tr, bdd init, bdd goal, bdd *sa,domdef *def) {
  bdd acc = goal;               // acc encoded using new state vars (visited states)
  bdd oldacc = bddfalse;        // old acc
  bdd preimage, prunedpreimage; 
  int i,nc;

  // make universal plan by backward chaining from goal
  i = 0; // preimage count
  
  //main loop
  while (acc != oldacc)
    {
      if ( (acc & init) == init ) 
	return 1;
      preimage = strongpreimage(tr,acc,def); // act*current bdd
      prunedpreimage = prunestates(preimage,acc); // act*current bdd

      if (prunedpreimage != bddfalse)
	{
	  *sa |= prunedpreimage;
	  oldacc = acc;
	  acc |= projact(prunedpreimage,def);
	  if (verbosity > 0)
	    cout << "No " << i << " Added a strong preimage\n";
	  i++;
	}
      else
	{
	  faircycles(tr,&acc,sa,&oldacc,def);
	  if (verbosity > 0)
	    cout << "No " << i << " Added a fair cycle preimage\n";
	  i++;
	}
    }
  return 0;
}




/****************************************************** 
 *          OP  
 *          Optimistic planning algorithm
 ******************************************************/


// IN
//  tr     : Partitioned transition relation
//  init   : Initial state (encoded in current state vars)
//  goal   : Goal state    (encoded in current state vars)
//  sa     : plan in act * current-state 
//  def    : Domain definition 
// OUT
//  plan ignoring nondeterminism (uses the real preimage in back chaining)
//  1:  init subset of SA (plan found)
//  0:  init not subset of SA (no plan found)
int OP(trel *tr, bdd init, bdd goal, bdd *sa,domdef *def) {
  bdd acc = goal;               // acc encoded using new state vars (visited states)
  bdd preimage, prunedpreimage; 
  int i = 0;
  
  // make universal plan by backward chaining from goal
  while ( (acc & init) != init ) 
    {
      preimage = weakpreimage(tr,acc,def); // act*current bdd
      prunedpreimage = prunestates(preimage,acc); // act*current bdd

      if (prunedpreimage != bddfalse)
	{
	  *sa |= prunedpreimage;
	  acc |= projact(prunedpreimage,def);
	  if (verbosity > 0)
	    cout << "Added OP preimage " << i++ << endl;
	}
      else
	// fix point found, but init not subset of acc
	return 0;
    }
  // init subset of acc
  return 1;
}



/****************************************************** 
 *    Aux. functions for game planning algorithms 
 ******************************************************/


// IN
//  sa     : state action set (sysact * current state)
//  def    : Domain definition (containing a description of the 
//           location of variables in bdds.)
// OUT
//  states of sa (current state)
bdd states(bdd sa,domdef *def)  {
  return bdd_exist(sa, def->rels->sysactions); 
}


// IN
//  tr     : Partitioned transition relation
//  v      : set of vistied states (current state)
//  def    : Domain definition (containing a description of the 
//           location of variables in bdds.)
// OUT
// Jv(s,ae) : environment actions ae belonging to a joint action
//            leading from s to s', where is in the set of states V.
// Jv(s,ae) = Eas.As'.T(s,as,ae,s')
bdd J(trel *tr, bdd v,domdef *def) {
  int i;
  bdd res;

  res = bdd_replace(v,def->rels->old2new);
  for (i=0; i<tr->partnum; i++)
    {
      res = bdd_appex(res,tr->p[i].exp,bddop_and,tr->p[i].varset);
      if (verbosity > 1)
	cout << "J finished partition " << i << " node count: " 
	     << bdd_nodecount(res) << "\n";
    }

  res = bdd_exist(res, def->rels->sysactions);

  return res;
}


// IN
//  tr     : Partitioned transition relation
//  def    : Domain definition (containing a description of the 
//           location of variables in bdds.)
// OUT
//  A(s,ae) : environment actions applicable in s
//  A(s,ae) = Eas.Es'.T(a,as,ae,s')
bdd A(trel *tr,domdef *def) {
  
  bdd res = bddtrue;
  int i;
  
  for (i=0; i<tr->partnum; i++)
    {
      res = bdd_appex(res,tr->p[i].exp,bddop_and,tr->p[i].varset);
      if (verbosity > 1)
	cout << "A finished partition " << i << " node count: " 
	     << bdd_nodecount(res) << "\n";
    }
  res = bdd_exist(res, def->rels->sysactions);
  return res;
}





// IN
//  tr     : Partitioned transition relation
//  A      : constant A expression (applicable env acts)
//  v      : set of vistied states (current state)
//  def    : Domain definition (containing a description of the 
//           location of variables in bdds.
// OUT
//  Game Preimage in (act * current state)
//  GPreImage(s,as) = (Aae. A(s,ae) => Jv(s,ae)) /\
//                    Eae.Es'.T(s,as,ae,s') /\ V(s') /\ ~V(s) 
//  (obs: states in V are pruned from the preimage, by ~V(s))
bdd GPreImage(trel *tr, bdd A,  bdd v,domdef *def) {
  int i;
  bdd res;

  res =  bdd_replace(v,def->rels->old2new);
  res &= !v;
  for (i=0; i<tr->partnum; i++)
    {
      res = bdd_appex(res,tr->p[i].exp,bddop_and,tr->p[i].varset);
      if (verbosity > 1)
	cout << "GPreImage finished partition " << i << " node count: " 
	     << bdd_nodecount(res) << "\n";
    } 
  res = bdd_exist(res,def->rels->envactions); 
  res &= bdd_forall(A >> J(tr,v,def),def->rels->envactions);
  return res; 
}


/****************************************************** 
 *          OGP  
 *          Optimistic Game planning algorithm
 ******************************************************/




// IN
//  tr     : Partitioned transition relation
//  init   : Initial state (encoded in current state vars)
//  goal   : Goal state (encoded in current state vars)
//  sa     : plan to return in act * current-state (indirect output)
//  def    : Domain definition (containing a description of the 
//           location of variables in bdds.
//  outfile: Name of output file
// OUT
//  optimistic game plan
//  1:  init subset of SA (plan found)
//  0:  init not subset of SA (no plan found)
int OGP(trel *tr, bdd init, bdd goal, bdd *sa,domdef *def) {
  bdd sap;           // precomponent          (act*current state)
  bdd v = goal;      // set of visited states (current state)
  bdd a;             // constant A expression used in GPreImage
  int i = 0;

  a = A(tr,def);
  while ( (v & init) != init ) 
    {
      sap =  GPreImage(tr,a,v,def);
      if (sap == bddfalse)
	return 0; // fix point found, but init not subset of sa
      else
	{
	  *sa |= sap;
	  v |= states(*sa,def);
	  if (verbosity > 0)
	    cout << "Added AOP " << i++ << endl;
	}
    }
  // init subset of as
  return 1;
}



/****************************************************** 
 *          SCGP  
 *          Strong Cyclic Game planning algorithm
 ******************************************************/


// IN
//  tr     : Partitioned transition relation
//  v      : set of vistied states and states in precomponent 
//  as     : gamepreimages 
//  def    : Domain definition (containing a description of the 
//           location of variables in bdds.
// OUT
//  as pruned for sa pairs that can lead out of v
bdd pruneoutgoing(trel *tr,bdd v,bdd sa,domdef *def) {
  bdd res;
  int i;

  res = !bdd_replace(v,def->rels->old2new); // next state not allowed to be in v

  for (i=0; i < tr->partnum; i++)
    {
      res = bdd_appex(res,tr->p[i].exp,bddop_and,tr->p[i].varset);
      if (verbosity > 1)
	cout << "Pruneoutgoing finished partition " << i << " node count: " 
	     << bdd_nodecount(res) << "\n";
    }
  res = bdd_exist(res, def->rels->envactions);

    return !res & sa; // constrain as to a*s pairs that cannot lead out of v 
}



// IN
//  tr     : Partitioned transition relation
//  v      : set of vistied states and states in precomponent up to as (current state)
//  as     : gamepreimages 
//  def    : Domain definition (containing a description of the 
//           location of variables in bdds.
// OUT
//  as pruned for nongame states wrt. v 
bdd prunenotgame(trel *tr,bdd A,bdd v,bdd sa,domdef *def) {
  bdd j;
  bdd gs; // game states wrt. v in sa
  int i;

  // find game states wrt. v for T reduces to the state actions pairs in SA
  j = sa;
  for (i=0; i<tr->partnum; i++) // compute: T /\ as
    {
      j = bdd_appex(j,tr->p[i].exp,bddop_and,tr->p[i].varset);
      if (verbosity > 1)
	cout << "Prunenotgame finished partition " << i << " node count: " 
	     << bdd_nodecount(j) << "\n";
    }
  j = bdd_exist(j, def->rels->sysactions);

  // game states in sa
  gs = bdd_forall(A >> j,def->rels->envactions);
  return sa & gs;
}






// IN
//  tr     : Partitioned transition relation
//  A      : constant A expression (applicable env acts)
//  v      : set of visitied states (current state)
//  sa     : gamepreimages in current precomponent
//  i      : number of gamepreimages in current precomponent
//  def    : Domain definition (containing a description of the 
//           location of variables in bdds.
// OUT
//  Pruned Game Precomponent in (act * current state)
bdd PruneSCG(trel *tr, bdd A, bdd v, bdd *SaOrg, int sanum, domdef *def) {
  int i;
  bdd sa[MAXPLANLENGTH]; // gamepreimages in component
  bdd sao;
  bdd sap;
  bdd vc;    // all vistied states and states in sa[0] - sa[i];
  bdd vt;    // all visited states and states in the current precomponent


  // copy sa
  for (i=0; i<sanum; i++)
    sa[i] = SaOrg[i];
    
  sap = bddfalse;  
  sao = bddfalse;
  vc = v;
  vt = v;
  for (i=0; i<sanum; i++)
    vt |= states(sa[i],def); 

  for (i=0; i<sanum; i++)
    {
      sao |= sa[i];
      sa[i] = pruneoutgoing(tr,vt,sa[i],def);
      sa[i] = prunenotgame(tr,A,vc,sa[i],def);
      vc = states(sa[i],def);
      sap |= sa[i];
    }
  
  while (sap != sao)
    {
      sao = sap;
      vc = v;
      vt = v;
      for (i=0; i<sanum; i++)
	vt |= states(sa[i],def); 

      for (i=0; i<sanum; i++)
	{
	  sa[i] = pruneoutgoing(tr,v,sa[i],def);
	  sa[i] = prunenotgame(tr,A,vc,sa[i],def);
	  vc |= states(sa[i],def);
	  sap |= sa[i];
	}
    }
  return sap;
}

// IN
//  tr     : Partitioned transition relation
//  A      : constant A expression (applicable env acts)
//  v      : set of vistied states (current state)
//  def    : Domain definition (containing a description of the 
//           location of variables in bdds.
// OUT
//  Game Precomponent in (act * current state)
bdd PreCompSCGP(trel *tr, bdd A,bdd v,domdef *def) {
  int i;
  bdd sac;               // current candidate component (sysact * current state)
  bdd sa[MAXPLANLENGTH]; // gamepreimages in component
  bdd vc;                // currently visited states 

  i = 0;
  vc = v;
  sa[0] = GPreImage(tr,A,vc,def);
  while (sa[i] != bddfalse)
    {
      sac = PruneSCG(tr,A,v,sa,i+1,def);
      if (sac != bddfalse) 
	return sac;
      else
	{
	  vc |= states(sa[i],def);
	  i++;
	  sa[i] = GPreImage(tr,A,vc,def);
	  if (verbosity > 0)
	    cout << "  Added SCA layer " << i << " (with " << bdd_nodecount(sa[i]) << " nodes) " << " to SCA precomponent\n";
	}
    }
  return bddfalse; // fixpoint found that does not contain a game precomponent
}



// IN
//  tr     : Partitioned transition relation
//  init   : Initial state (encoded in current state vars)
//  goal   : Goal state (encoded in current state vars)
//  sa     : plan to return in act * current-state (indirect output)
//  def    : Domain definition (containing a description of the 
//           location of variables in bdds.
//  outfile: Name of output file
// OUT
//  optimistic game plan
//  1:  init subset of SA (plan found)
//  0:  init not subset of SA (no plan found)
int SCGP(trel *tr, bdd init, bdd goal, bdd *sa,domdef *def) {
  bdd sap;           // precomponent          (act*current state)
  bdd v = goal;      // set of visited states (current state)
  bdd a;             // constant A expression used in GPreImage
  int i = 0;

  a = A(tr,def);
  while ( (v & init) != init ) 
    {
      sap =  PreCompSCGP(tr,a,v,def);
      if (sap == bddfalse)
	return 0; // fix point found, but init not subset of sa
      else
	{
	  *sa |= sap;
	  v |= states(*sa,def);
	  if (verbosity > 0)
	    cout << "Added SCA precomponent " << i++ << endl;	      
	}
    }
  // init subset of sa
  return 1;
}
















